--[[

    EnhaPrio by Stormcast @ Lightbringer EU
    Simple addon to display Enhance Shamans spellcasting/abilities priorities
    depending on the situation.
    Inspired and based loosely on ShamWow_Enhance

    Legend:

    SS - Stormstrike
    LB - Lightning Bolt
    FS - Flame Shock
    LS - Lightning Shield
    MT - Magma Totem
    ES - Earth Shock
    LL - Lava Lash
    MS - Maelstrom
    WF - Windfury weapon
    FT - Flame Tongue
    SR - Shamanistic Rage
    FN - Fire Nova
    FE - Fire Elemental Totem
    SR - Shamanistic Rage
    LA - Lava Burst
    UE - Unleash Elements
    ST - Searing Totem
    SF - Searing Flames
    GHW - Greater Healing Wave
    
    LB5 - Lightning Bolt used when 5 stacks of Maelstrom Weapon are up
    LB4 - Lightning Bolt used when 4 stacks of Maelstrom Weapon are up
    LB3 - Lightning Bolt used when 3 stacks of Maelstrom Weapon are up and next cd is more than 1.5 sec away
    ES9 - Earth Shock when 7-9 stacks of Lightning Shield are up
    LAX - Lava Burst when there is less than 3 stacks of MW and next cd is more than 2 sec away

    This addon doesn't take long cooldown abilities like Feral Spirit or
    Fire Elemental into account. Usage of those skills is up to you (they
    are still an important part of enhancement shamans skillset during bosses).

    Shamanistic Rage will be suggested when you are low on mana. And if you
    are missing weapon buffs, they will be suggested too.

Below is the priority. You can change the order of these, if you think it necessary
or comment out with "--" if not wanted. ]]
Priority = {
	Enhancement = { -- priorities used in enhancement spec
		"GHW", -- Greater Healing Wave, when in critical health
		"WF", -- weapon buffs (windfury)
		"FT", -- weapon buffs (flametongue)
		
		-- actual damage spells

		"ST0", -- Searing Totem
		"SS", -- Stormstrike
		"LL", -- Lava Lash
		"LB5", -- Lightning Bolt if there are 5 Maelstrom stacks
		"FS", -- Flame Shock if there's unleash flame buff
		"UE", -- Unleash elements
		--"LB4", -- lightning bolt with 4 charges
		"ES", -- Earth Shock
		--"LB3", -- lightning bolt with 3 charges and time to spare
		--"LAX", -- Lava Burst with conditions (for enha)

		"LS", -- Lightning Shield if it isn't active on you
	},
	Elemental = { -- priorities used in elemental spec
		"FT", -- weapon buffs (flametongue)
	
		-- actual damage spells
		
		"ST", -- searing totem
	 	"FS", -- Flame Shock if there's less than 1.5 sec left on the dot
	 	"LA", -- lava burst
	 	"ES9", -- earth shock when there is 9 stacks of ls
	 	"LS", -- lightning shield, if you don't have one
	 	"LB"  -- lightning bolt
	}
}


--[[ Ok, don't mess with anything below this (unless you know what you're doing) ]]--


-- initialize the variables (make them globals)
local EnhaPrio = LibStub( "AceAddon-3.0" ):NewAddon( "EnhaPrio", "AceConsole-3.0", "AceEvent-3.0" );
local LBF = LibStub("Masque", true);
if LBF then
    Group = LBF:Group("EnhaPrio");
end
local mainFrame, target, skins;
local maxQueueSize = 8;
local queue = {};
local cdqueue = {};
hasLS = false;
hasMT = false;
melee = false;
ranged = false;
fsLeft = 0;
noSS = false;
noLS = false;
hostile = false;
hasUF = false;
--lowMana = false;
hasMH = false;
hasOH = false;
timeLeft = 0;
mwAmount = 0;
hasTotem = false;
mode = nil;
lsStack = 0;
health = 100;
sfAmount = 0;
nextCD = 0;

-- configs (default values)
local defaults = {
	char = {
		x = 0,
		y = 0,
		relativePoint = "CENTER",
		size = 50,
		spacing = 0,
		locked = false,
		clickthru = false,
		updateFrequency = .2,
		--manaThreshold = 20,
		maxQueue = 4,
		displayMW = true,
		queueDirection = "RIGHT",
		displayGCD = true,
		enableLongCD = true,
		sizeFactor = 1,
		alwaysShow = "NO",
		healthLevel = 20,
		enhancement = true,
		elemental = true
	}
};


-- the spells we'll be using
local Spells = { 
	SS = GetSpellInfo(17364), -- stormstrike
	LL = GetSpellInfo(60103), -- lava lash
	ES = GetSpellInfo(8042), -- earth shock
	MS = GetSpellInfo(51530), -- maelstrom weapon
	WF = GetSpellInfo(8232), -- windfury weapon
	FT = GetSpellInfo(8024), -- flametongue weapon
	LB = GetSpellInfo(403), -- lightning bolt
	LS = GetSpellInfo(324), -- lightning shield
	SR = GetSpellInfo(30823), -- shamistic rage
	FR = GetSpellInfo(51533), -- feral spirit
	FE = GetSpellInfo(2894), -- fire elemental totem
	FS = GetSpellInfo(8050), -- flame shock
	MT = GetSpellInfo(8190), -- magma totem
	FN = GetSpellInfo(1535), -- fire nova
	ST = GetSpellInfo(3599), -- searing totem
	LA = GetSpellInfo(51505), -- Lava Burst
	EM = GetSpellInfo(16166), -- Elemental Mastery
	UE = GetSpellInfo(73680),  -- Unleash Elements
	UF = GetSpellInfo(73683), -- Unleash Flame
	GHW = GetSpellInfo(77472), -- Greater Healing Wave
	SF = GetSpellInfo(77661) -- Searing Flames, the debuff from searing totem
}

-- here are the different actions (adding stuff to queue according to the situation)
local Actions = {

	WF = function ()
		if not hasMH then
			addToQueue(Spells.WF);
		end
	end,
	
	FT = function ()
		if (mode == "Enhancement" and not hasOH) or (mode == "Elemental" and not hasMH) then
			addToQueue(Spells.FT);
		end
	end,
	
	--[[
	SR = function ()
		-- do shamanistic rage
		if isCastable(Spells.SR.name) and lowMana and melee then
			addToQueue(Spells.SR.name);
		end
	end,
	]]
	
	UE = function ()
	    if isCastable(Spells.UE) then
			addToQueue(Spells.UE);
		end
	end,

	LB5 = function ()
		-- do lb, if 5 buffs
		if mwAmount == 5 and ranged and health > EnhaPrio.db.char.healthLevel then
			addToQueue(Spells.LB);
		end
	end,
	
	LB4 = function ()
		-- do lb, if 4 buffs
		if mwAmount == 4 and ranged and health > EnhaPrio.db.char.healthLevel then
			addToQueue(Spells.LB);
		end
	end,
	
	LB3 = function ()
		-- do lb, if 3 buffs and more then 1.5 sec for the next cd to come out
		if mwAmount == 3 and ranged and health > EnhaPrio.db.char.healthLevel and nextCD > 1.5 then
			addToQueue(Spells.LB);
		end
	end,
	
 	LB = function ()
		-- just lightning bolt (for elemental)
  		if isCastable(Spells.LB) then
			addToQueue(Spells.LB);
		end
	end,

	GHW = function ()
	    -- do greater healing wave, if has 5 buffs and health is low
	    if mwAmount == 5 and health <= EnhaPrio.db.char.healthLevel and EnhaPrio.db.char.healthLevel > 0 then
		addToQueue(Spells.GHW);
	    end
	end,
	FS = function ()
		-- if there is under 1.5sec left on flame shock on the target
		-- if isCastable(Spells.FS) and ranged and fsLeft <= 3 then
		
		if isCastable(Spells.FS) and ranged and (hasUF or (mode == "Elemental" and fsLeft <= 3)) then -- changed
			addToQueue(Spells.FS);
		end
	end,
	
	LS = function ()
		-- lightningshield
		if noLS then
			addToQueue(Spells.LS);
		end
	end,
	
 	ST0 = function ()
		-- searing totem, with no stacks up
		if sfAmount == 0 and melee and not hasTotem then
			addToQueue(Spells.ST);
		end
	end,
	
	STX = function ()
		-- searing totem, with some stacks up
		if sfAmount > 0 and sfAmount < 5 and melee and not hasTotem then
			addToQueue(Spells.ST);
		end
	end,
	
	ST = function ()
		-- searing totem, if you don't have the totem down, for elemental
		if not hasTotem then
			addToQueue(Spells.ST);
		end
	end, 
	
	ES = function ()
		-- earth shock
		if isCastable(Spells.ES) and ranged and not hasUF then
			addToQueue(Spells.ES);
		end
	end,
	
	ES9 = function ()
		-- earth shock when there is 7-9 stacks on ls
		if isCastable(Spells.ES) and ranged and fsLeft > 3 and lsStack >= 7 then
			addToQueue(Spells.ES);
		end
	end,
	
	SS = function () 
		-- Stormstrike
		if isCastable(Spells.SS) and melee then
			addToQueue(Spells.SS);
		end
	end,
	
	LL = function ()
		-- lava lash
		if isCastable(Spells.LL) and melee then
			addToQueue(Spells.LL);
		end	
	end,
	
	LA = function ()
		-- lava burst
		if isCastable(Spells.LA) and ranged and fsLeft > 2 then
			addToQueue(Spells.LA)
		end
	end,
	
	LAX = function ()
		-- lava burst for enha use
		if isCastable(Spells.LA) and mwAmount < 3 and nextCD > 2 then
			addToQueue(Spells.LA)
		end
	end
}

local Cooldowns = {
	Enhancement = {	
		"FS", "LS", "ES", "SS", "LL", "UE"  --, "ST"
	},
	Elemental = {
		"ES", "LA"
	}
}


-- makes a queue of cooldowns
function makeCDQueue()

	local cdq = {};
	for i, n in pairs(Cooldowns[mode]) do
		if getCD(n) > 0 then
		    table.insert(cdq, n);
		end
	end
	local ssDebuff = UnitDebuff("target", "Stormstrike", _, "PLAYER");


	local sorter = function (a, b)
		-- if the difference between cooldowns is less than a second,
		-- treat them as having the same cd (so priority should come
		-- from the priority list)
		if math.abs(getCD(a) - getCD(b)) < 0.5 then
		  	for i, v in ipairs(Priority[mode]) do
		  	
		  	    if v == "ES9" then v = "ES" end
		  		if v == a then av = i end
		  		if v == b then bv = i end
		  	end
		  	return av < bv;
		else
			return getCD(a) < getCD(b);
		end
	end
	
	table.sort(cdq, sorter);
	cdqueue = {};
	for i, n in ipairs(cdq) do
		table.insert(cdqueue, Spells[n]);
	end
end

-- get the time when the spell should be cast again (used for cd queue)
function getCD(n)
	local start, duration;
	
	-- gcd control
	gcdstart, gcdduration = GetSpellCooldown(Spells.LB);
	gcdtime = gcdstart + gcdduration;
	
	if n == "FE" or n == "ST" then
		_, _, start, duration = GetTotemInfo(1);
	else
		start, duration = GetSpellCooldown(Spells[n]);
	end
	
	local left = start + duration - GetTime();
	
	if mode == "Elemental" then
		if n == "ES" and left >= (fsLeft - 3) then
			duration = 0;
		end
		if n == "FS" and left < (fsLeft - 3) then
			duration = 0;
		end
	elseif mode == "Enhancement" then
		if n == "ES" and hasUF then
			duration = 0;
		end
		if n == "FS" and not hasUF then
			duration = 0;
		end
	end
	
	if n == "FN" and not hasMT then
		duration = 0;
	end
	
	if duration > 0 and (start + duration) > gcdtime then
		return start + duration;
	else
		return 0;
	end
end


-- can you cast that spell
function isCastable(spellName)
	-- check if you can cast that spell in one gcd
	local _, GCD = GetSpellCooldown(Spells.LB);
	local _, duration = GetSpellCooldown(spellName);
	return duration == GCD or duration == 0;
end

-- add a spell to the queue
function addToQueue(spell)
	queue[#queue+1] = spell;
end

-- round a number (odd that lua doesn't have that in standard classes)
function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end


-- refreshes the queue according to the priorities
-- check stuff and then run the queue
function refreshQueue()

	-- players health percentage
	health = UnitHealth("player") / (UnitHealthMax("player") / 100);
	
	-- players buffs (maelstrom, lightning shield)
	noLS = true;
	hasUF = false;
	mwAmount = 0;
	for i=1,40 do
		local name, _, _, count = UnitBuff("player", i);
		if not name then
			break; -- end of buffs
		end
		if name == Spells.MS then
		    mwAmount = count;
		elseif name == Spells.LS then
			lsStack = count;
			noLS = false;
		elseif name == Spells.UF then
		    hasUF = true;
		end
	end
	
	-- targets debuffs (fire shock)
	noSS = true;
	fsLeft = 0;
	sfAmount = 0;
	for i=1,40 do
		local name, _, _, dcount, _, _, expirationTime, caster = UnitDebuff("target", i);
		if not name then
			break; -- end of debuffs
		end
		if caster == "player" and name == Spells.FS then
			fsLeft = expirationTime - GetTime();
		elseif caster == "player" and name == Spells.SS then
			noSS = false;
		elseif caster == "player" and name == Spells.SF then
			sfAmount = dcount;
		end
	end
	
	-- totems (magma or elemental)
	local _, totemName = GetTotemInfo(1);	
	if totemName == Spells.MT or totemName == Spells.FE then
		hasMT = true;
	else
		hasMT = false;
	end
	
	local _, totemName = GetTotemInfo(1);
	if totemName == "" then
	    hasTotem = false;
	else
	    hasTotem = true;
	end
	
	-- weapon buffs
	hasMH, _, _, hasOH = GetWeaponEnchantInfo();
	
	-- ranges
	melee = IsSpellInRange(Spells.LB, 'target') == 1; -- if you are in range of melee attacks (using flame shock here too... )
	ranged = IsSpellInRange(Spells.LB, 'target') == 1; -- if you are in range of flame shock
	
	-- if we want to show them always
	if EnhaPrio.db.char.alwaysShow == "YES" then
	    melee, ranged = true, true;
	end

  	-- now loop through the actions
	for i, v in ipairs(Priority[mode]) do
		Actions[v]();
	end
end

function EnhaPrio:reCalculate()
	-- check if the target is hostile and you are not mounted or on a vehicle or dead
	queue = {};
	hostile = UnitName("target") and UnitCanAttack("player","target") and UnitHealth("target") > 0;
	mounted = false; --- make it work!!!
	dead = UnitHealth("target") < 1;
	yourdead = UnitHealth("player") < 1;
	local maxQueueSize = self.db.char.maxQueue;
	
	if self.db.char.alwaysShow == "YES" or self.db.char.alwaysShow == "BUFFS" then
	    hostile = true;
	end

	if hostile and not yourdead and mode ~= nil then
	    timeLeft = self.db.char.updateFrequency;
		refreshQueue();
		makeCDQueue();
	
	    -- show maelstrom wep
	    if mode == "Enhancement" and self.db.char.displayMW and mwAmount > 0 and ranged then
	        mainFrame.text:SetText(mwAmount);
	        if mwAmount == 1 then
	        	mainFrame.text:SetTextColor(1, 1, 1, 1);
			elseif mwAmount == 2 then
			    mainFrame.text:SetTextColor(1, 1, 0, 1);
			elseif mwAmount == 3 then
			    mainFrame.text:SetTextColor(1, 1, 0, 1);
			elseif mwAmount == 4 then
			    mainFrame.text:SetTextColor(1, .5, 0, 1);
			elseif mwAmount == 5 then
			    mainFrame.text:SetTextColor(1, 0, 0, 1);
			end
		elseif mode == "Elemental" and self.db.char.displayMW and lsStack > 0 and ranged then
		    mainFrame.text:SetText(lsStack);
	        if lsStack > 0 and lsStack < 6 then
	        	mainFrame.text:SetTextColor(1, 1, 1, 1);
			elseif lsStack == 6 then
			    mainFrame.text:SetTextColor(1, 1, 0, 1);
			elseif lsStack == 7 then
			    mainFrame.text:SetTextColor(1, 1, 0, 1);
			elseif lsStack == 8 then
			    mainFrame.text:SetTextColor(1, .5, 0, 1);
			elseif lsStack == 9 then
			    mainFrame.text:SetTextColor(1, 0, 0, 1);
			end
		else
		    mainFrame.text:SetText("");
	    end
	    
	    -- disable this for now
	    mainFrame.rage.texture:SetTexture(nil);
		mainFrame.rage:Hide();
	    
	    -- wolf/em and ele
	    if self.db.char.enableLongCD then
	    	if mode == "Enhancement" then
		    	if isCastable(mainFrame.wolf.id) and ranged then
		    		mainFrame.wolf.texture:SetTexture(GetSpellTexture(Spells.FR));
					mainFrame.wolf:Show();
				else
					mainFrame.wolf.texture:SetTexture(nil);
					mainFrame.wolf:Hide();
				end
				--[[
				if isCastable(mainFrame.rage.id) and ranged then
		    		mainFrame.rage.texture:SetTexture(GetSpellTexture(Spells.SR));
					mainFrame.rage:Show();
				else
					mainFrame.rage.texture:SetTexture(nil);
					mainFrame.rage:Hide();
				end
				]]
			elseif mode == "Elemental" then
			    if isCastable(Spells.EM) and ranged then
		    		mainFrame.wolf.texture:SetTexture(GetSpellTexture(Spells.EM));
					mainFrame.wolf:Show();
				else
					mainFrame.wolf.texture:SetTexture(nil);
					mainFrame.wolf:Hide();
				end
				mainFrame.rage.texture:SetTexture(nil);
				mainFrame.rage:Hide();
			end
			if isCastable(mainFrame.elemental.id) and ranged then
				mainFrame.elemental.texture:SetTexture(GetSpellTexture(Spells.FE));
				mainFrame.elemental:Show();
			else
				mainFrame.elemental.texture:SetTexture(nil);
				mainFrame.elemental:Hide();
			end
	    end
	    
	    -- try to figure out the next cd
	    local nextSpell = cdqueue[1];
	    if nextSpell then
			local start, duration = GetSpellCooldown(nextSpell);
			nextCD = start + duration - GetTime();
		end

	    -- merge the normal and cd queues
	    local tillcd = #queue;
	    for i, n in ipairs(cdqueue) do table.insert(queue, n) end

		-- let's draw the queue
		for i=1, maxQueueSize do
			local spell = queue[i];
			local f = spellQueueFrames[i];
			f.spell = spell;
			if spell then
			
				if i <= tillcd then
					-- normal skills
					f.spellTexture:SetTexture(GetSpellTexture(spell));
					check(f, spell);
					if i > 1 then CooldownFrame_SetTimer(f.cooldown, 0, 0, 0) end
					if i == 1 and spell == Spells.LB then self:SPELL_UPDATE_COOLDOWN() end
					f.cooldownText:SetText("");
					f:SetAlpha(1);
	    			f:Show();
    			else
    				-- normal queue ended, going on cooldowns
					f.spellTexture:SetTexture(GetSpellTexture(spell));
					local start, duration;
					if spell == Spells.ST then
						_, _, start, duration = GetTotemInfo(1);
					else
						start, duration = GetSpellCooldown(spell);
					end
					check(f, spell);
				    CooldownFrame_SetTimer(f.cooldown, start, duration, 1);
				    local left = round(start + duration - GetTime());
				    if left > 0 then f.cooldownText:SetText(left) else f.cooldownText:SetText("") end
				    f:SetAlpha(0.7);
					f:Show();
				end
    			
			else
				-- nothing left in the queues
				f.spellTexture:SetTexture(nil);
				f.cooldownText:SetText("");
				f:SetAlpha(1);
    			f:Hide();
			end
		end
	else
		-- clear the icons
		for i=maxQueueSize,1,-1 do
			local spell = queue[i];
			local f = spellQueueFrames[i];
			f.spellTexture:SetTexture(nil);
			f:Hide();
		end
		
		mainFrame.text:SetText("");
		mainFrame.wolf:Hide();
		mainFrame.elemental:Hide();
		mainFrame.rage:Hide();
		
	end
end

-- check for out of mana or out of range
function check(f, spell)
	name, _, _, cost = GetSpellInfo(spell);
	if IsSpellInRange(spell, 'target') == 0 then
	    f.spellTexture:SetVertexColor(1, 0, 0);
	elseif cost and UnitPower('player') < cost then
	    f.spellTexture:SetVertexColor(0.4, 0.4, 0.4);
	else
	    f.spellTexture:SetVertexColor(1, 1, 1);
	end
end

-- functions to take care of the addon (visuals and saving etc)
function EnhaPrio:SaveLocation()
	point, relativeTo, relativePoint, xOfs, yOfs = mainFrame:GetPoint();
	self.db.char.x = xOfs;
	self.db.char.y = yOfs;
	self.db.char.relativePoint = relativePoint;
end

function EnhaPrio:RepositionFrames(queue)
    local maxQueueSize = self.db.char.maxQueue;
    local spacing = self.db.char.spacing;

	if queue then
	    for i, f in ipairs(spellQueueFrames) do
	  		f:Hide();
		end
		mainFrame.wolf:Hide();
		mainFrame.elemental:Hide();
		mainFrame.rage:Hide();
	end
	
	-- text
	mainFrame.text:SetTextHeight(self.db.char.size / 2);
	mainFrame.text:ClearAllPoints();
	if self.db.char.queueDirection == "LEFT" then
		mainFrame.text:SetPoint("LEFT", mainFrame, "RIGHT", spacing, 0);
	else
	    mainFrame.text:SetPoint("RIGHT", mainFrame, "LEFT", (spacing * -1), 0);
	end
    	
	-- wolf, rage and elemental frames
	mainFrame.wolf:SetWidth(self.db.char.size / 2)
	mainFrame.wolf:SetHeight(self.db.char.size / 2)
	mainFrame.wolf:ClearAllPoints();
	mainFrame.elemental:SetWidth(self.db.char.size / 2)
	mainFrame.elemental:SetHeight(self.db.char.size / 2)
	mainFrame.elemental:ClearAllPoints();
	mainFrame.rage:SetWidth(self.db.char.size / 2)
	mainFrame.rage:SetHeight(self.db.char.size / 2)
	mainFrame.rage:ClearAllPoints();
	if self.db.char.queueDirection == "DOWN" or self.db.char.queueDirection == "UP" then
		mainFrame.wolf:SetPoint("TOPLEFT", mainFrame, "TOPRIGHT", spacing, 0);
	    mainFrame.elemental:SetPoint("BOTTOMLEFT", mainFrame, "BOTTOMRIGHT", spacing, 0);
	else
	    mainFrame.wolf:SetPoint("TOPLEFT", mainFrame, "BOTTOMLEFT", 0, (spacing * -1));
	    mainFrame.elemental:SetPoint("TOPRIGHT", mainFrame, "BOTTOMRIGHT", 0, (spacing * -1));
	end
	
    if self.db.char.queueDirection == "DOWN" then
	  	mainFrame.rage:SetPoint("TOPLEFT", mainFrame, "BOTTOMRIGHT", spacing, 0);
	elseif self.db.char.queueDirection == "UP" then
		mainFrame.rage:SetPoint("BOTTOMLEFT", mainFrame, "TOPRIGHT", spacing, 0);
	elseif self.db.char.queueDirection == "LEFT" then
		mainFrame.rage:SetPoint("TOPRIGHT", mainFrame, "BOTTOMLEFT", 0, (spacing * -1));
	elseif self.db.char.queueDirection == "RIGHT" then
		mainFrame.rage:SetPoint("TOPLEFT", mainFrame, "BOTTOMRIGHT", 0, (spacing * -1));
	end	
    	
    	
	-- buttons
	for i=1,maxQueueSize do
		local f = spellQueueFrames[i];
		local size = self.db.char.size;
		if i > 1 then
			size = self.db.char.sizeFactor * size;
		end
		f:SetWidth(size)
		f:SetHeight(size)
		f:ClearAllPoints();
		if i == 1 then
			f:SetPoint("CENTER", mainFrame, "CENTER");
		else
		    if self.db.char.queueDirection == "RIGHT" then
		    	f:SetPoint("LEFT", spellQueueFrames[i-1], "RIGHT", spacing, 0);
            elseif self.db.char.queueDirection == "UP" then
			    f:SetPoint("BOTTOM", spellQueueFrames[i-1], "TOP", 0, spacing);
			elseif self.db.char.queueDirection == "LEFT" then
			    f:SetPoint("RIGHT", spellQueueFrames[i-1], "LEFT", (spacing * -1), 0);
            elseif self.db.char.queueDirection == "DOWN" then
			    f:SetPoint("TOP", spellQueueFrames[i-1], "BOTTOM", 0, (spacing * -1));
			end
		end
	end
	
	-- and reskin them
	if LBF then
		Group:ReSkin();
	end
end

-- print something to window
function swPrint(s)
    DEFAULT_CHAT_FRAME:AddMessage("EnhaPrio: ".. tostring(s));
end


--- on initialize
function EnhaPrio:OnInitialize()
	local AceConfigReg = LibStub("AceConfigRegistry-3.0")
	local AceConfigDialog = LibStub("AceConfigDialog-3.0")
	
	-- register shit
	self.db = LibStub("AceDB-3.0"):New("EnhaPrioDB", defaults, "char")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("EnhaPrio", self:GetOptions(), {"EnhaPrio", "fin"} )
	self.optionsFrame = AceConfigDialog:AddToBlizOptions("EnhaPrio","EnhaPrio")
	self.db:RegisterDefaults(defaults);
	
	-- Register for Function Events
	self:UnregisterAllEvents();
	--self:RegisterEvent("PLAYER_TARGET_CHANGED");
	self:RegisterEvent("SPELL_UPDATE_COOLDOWN");
	self:RegisterEvent("PLAYER_TALENT_UPDATE");

	-- create the main frame and configure it
	mainFrame = CreateFrame("Frame","EnhaPrioDisplayFrame",UIParent)
	mainFrame:SetFrameStrata("BACKGROUND")
	mainFrame:SetWidth(self.db.char.size)
	mainFrame:SetHeight(self.db.char.size)
	mainFrame:SetClampedToScreen(true)
	mainFrame:SetMovable(true)
	mainFrame:EnableMouse(true)

	-- add scripts for mouse events on the frame
	mainFrame:SetScript("OnMouseDown", function(self, button)
		if not EnhaPrio.db.char.locked and button == "LeftButton" then
			self:StartMoving();
		end
	end);
	mainFrame:SetScript("OnMouseUp", function(self) 
		self:StopMovingOrSizing(); 
		EnhaPrio:SaveLocation(); 
	end);
	mainFrame:SetScript("OnDragStop", function(self) 
		self:StopMovingOrSizing(); 
		EnhaPrio:SaveLocation(); 
	end);
	mainFrame:ClearAllPoints();
	mainFrame:SetPoint(self.db.char.relativePoint, self.db.char.x, self.db.char.y);

	-- let's create the individual frames for the icons
	spellQueueFrames = {};
	for i=maxQueueSize,1,-1 do
		local parentFrame = mainFrame;
		local f = CreateFrame("Button","EnhaPrioButton" .. i, parentFrame)
		f:SetFrameStrata("BACKGROUND")
		f:SetWidth(self.db.char.size)
		f:SetHeight(self.db.char.size)
		f:EnableMouse(false)
		f:SetMovable(false)
		f:SetClampedToScreen(true)
		f:ClearAllPoints();
		local spacing = self.db.char.spacing;
		f.spellTexture = f:CreateTexture(nil,"BACKGROUND");
		f.spellTexture:SetAllPoints(f);
		f.cooldown = CreateFrame("Cooldown", nil, f, "CooldownFrameTemplate");
		f.cooldown:SetAllPoints(f);
		f.cooldownText = f:CreateFontString(nil, "OVERLAY");
		f.cooldownText:SetFont("Fonts\\FRIZQT__.TTF", 42, "OUTLINE");
		f.cooldownText:SetTextHeight(self.db.char.size / 3);
		f.cooldownText:SetAllPoints();
		f.cooldownText:SetTextColor(1, 1, 1, 1);
		spellQueueFrames[i] = f;

        if LBF then Group:AddButton(f, {Icon = f.spellTexture, Cooldown = f.cooldown}) end
	end
	
	-- frames for wolf and elemental and rage
	mainFrame.wolf = CreateFrame("Button", "EnhaPrioFeralSpiritButton", mainFrame);
	mainFrame.wolf.id = 51533;
	mainFrame.wolf.texture = mainFrame.wolf:CreateTexture(nil,"BACKGROUND");
	mainFrame.wolf.texture:SetAllPoints(mainFrame.wolf);
	mainFrame.wolf:SetWidth(self.db.char.size / 2);
	mainFrame.wolf:SetHeight(self.db.char.size / 2);
	mainFrame.wolf:EnableMouse(false);
	mainFrame.wolf:SetMovable(false);
	mainFrame.wolf:SetClampedToScreen(true);
	mainFrame.wolf:ClearAllPoints();
	
	mainFrame.elemental = CreateFrame("Button", "EnhaPrioFireElementalButton", mainFrame);
	mainFrame.elemental.id = 2894;
	mainFrame.elemental.texture = mainFrame.elemental:CreateTexture(nil,"BACKGROUND");
	mainFrame.elemental.texture:SetAllPoints(mainFrame.elemental);
	mainFrame.elemental:SetWidth(self.db.char.size / 2);
	mainFrame.elemental:SetHeight(self.db.char.size / 2);
	mainFrame.elemental:EnableMouse(false);
	mainFrame.elemental:SetMovable(false);
	mainFrame.elemental:SetClampedToScreen(true);
	mainFrame.elemental:ClearAllPoints();
	
	mainFrame.rage = CreateFrame("Button", "EnhaPrioShamanisticRageButton", mainFrame);
	mainFrame.rage.id = 30823;
	mainFrame.rage.texture = mainFrame.rage:CreateTexture(nil,"BACKGROUND");
	mainFrame.rage.texture:SetAllPoints(mainFrame.rage);
	mainFrame.rage:SetWidth(self.db.char.size / 2);
	mainFrame.rage:SetHeight(self.db.char.size / 2);
	mainFrame.rage:EnableMouse(false);
	mainFrame.rage:SetMovable(false);
	mainFrame.rage:SetClampedToScreen(true);
	mainFrame.rage:ClearAllPoints();
	
	if LBF then 
		Group:AddButton(mainFrame.wolf, {Icon = mainFrame.wolf.texture});
		Group:AddButton(mainFrame.elemental, {Icon = mainFrame.elemental.texture}); 
		Group:AddButton(mainFrame.rage, {Icon = mainFrame.rage.texture});
	end
	
	
	-- text for maelstrom
	mainFrame.text = mainFrame:CreateFontString(nil,"OVERLAY")
	mainFrame.text:SetFont("Fonts\\FRIZQT__.TTF", 42, "THICKOUTLINE");
	mainFrame.text:SetTextHeight(self.db.char.size / 2);
	mainFrame.text:ClearAllPoints();
	mainFrame.text:SetTextColor(1, 1, 1, 1);
	
end

function EnhaPrio:OnEnable()
	local playerClass, englishClass = UnitClass("player");
	if UnitLevel('player') < 81 then
	    swPrint('Only characters lvl81 and above are supported.');
		mainFrame:Hide();
		return false;
	elseif englishClass ~= 'SHAMAN' then
		swPrint('You are not a shaman. Please reroll a shaman.');
		mainFrame:Hide();
		return false;
	else
		playerId = UnitGUID("player");
		self:ResolveSpec();

		-- Register chat commands.
		self:RegisterChatCommand("ep", function() self:OpenOptions() end);
		self:RegisterChatCommand("enhaprio", function() self:OpenOptions() end);
		
		-- move the frames side by side
		self:RepositionFrames();
		
		if self.db.char.locked and self.db.char.clickthru then
			mainFrame:EnableMouse(false)
		else
			mainFrame:EnableMouse(true)
		end
		 
		swPrint('Enabled');
	end
end

function EnhaPrio:Run()
    mainFrame:SetScript('OnUpdate', function(self, timeSinceLast)
		timeLeft = timeLeft - timeSinceLast;
		if timeLeft <= 0 then
			EnhaPrio:reCalculate();
		end
	end);
end

function EnhaPrio:Stop()
    mainFrame:SetScript('OnUpdate', nil);
end

function EnhaPrio:ResolveSpec()
	-- what spec are we using
	if GetSpellCooldown(Spells.SS) ~= nil and self.db.char.enhancement then
		mode = "Enhancement";
	elseif GetSpellCooldown(Spells.EM) ~= nil and self.db.char.elemental then
	 	mode = "Elemental";
	else
	    mode = nil;
	end
	
	if mode ~= nil then
		self:Run();
		mainFrame:Show();
	else
	    self:Stop();
	    mainFrame:Hide();
	end
end


-- :OpenOptions(): Opens the options window.
function EnhaPrio:OpenOptions()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrame);
end

function EnhaPrio:OnDisable()
	self:Stop();
	swPrint('Disabled');
end

function EnhaPrio:GetProperty(info)
	local propName = info[#info];
	local value = self.db.char[propName];
	return value;
end

function EnhaPrio:SetProperty(info, newValue)
	local propName = info[#info];
	self.db.char[propName] = newValue;
	if propName == 'spacing' or propName == 'size' or propName == 'queueDirection' or propName == 'displayMW' or propName == 'sizeFactor' then
		self:RepositionFrames();
	elseif propName == 'maxQueue' or propName == 'enableLongCD' then
	    self:RepositionFrames(true);
	elseif propName == 'enhancement' or propName == 'elemental' then
		self:ResolveSpec();
	end
	if mainFrame and propName == 'size' then
		mainFrame:SetWidth(self.db.char.size);
		mainFrame:SetHeight(self.db.char.size);
	end 
	
	-- trying click-through
	if self.db.char.locked and self.db.char.clickthru then
		mainFrame:EnableMouse(false)
	else
		mainFrame:EnableMouse(true)
	end
end

function EnhaPrio:PLAYER_TALENT_UPDATE(...)
   -- changed spec
   self:ResolveSpec();
   if mode ~= nil then
      self:Run();
      mainFrame:Show();
   end
end

function EnhaPrio:SPELL_UPDATE_COOLDOWN(...)
	if queue[1] then
	 	local GCDstart, GCD = GetSpellCooldown(Spells.LB);
	 	local start, duration = GetSpellCooldown(queue[1]);
	 	
	 	local GCDleft = GCDstart + GCD - GetTime();
	 	local left = start + duration - GetTime();
	 	
	 	if self.db.char.displayGCD and GCD == duration and mode == "Enhancement" then
	        CooldownFrame_SetTimer(spellQueueFrames[1].cooldown, GCDstart, GCD, 1);
	 	end
	end
end

function EnhaPrio:GetOptions()
	local options = { 
		name = "EnhaPrio",
		handler = EnhaPrio,
		type = 'group',
		childGroups ='tree',
		args = {

			locked = {
				type = "toggle",
				name = "Locked",
				get = "GetProperty",
				set = "SetProperty",
				order = 0,
			},
			clickthru = {
				type = "toggle",
				name = "Click-Through",
				desc = "Mouse clicks will go through the frame when it is locked.",
				get = "GetProperty",
				set = "SetProperty",
				order = 1,
			},

			enableGroup = {
			    type = "group",
			    name = "Enable specs",
			    order = 2,
			    inline = true,
			    args = {
                    enhancement = {
			            type = "toggle",
			            name = "Enhancement",
			            desc = "Enable EnhaPrio for enhancement spec.",
			            get = "GetProperty",
			            set = "SetProperty",
			            order = 1
					},
					elemental = {
					    type = "toggle",
					    name = "Elemental",
					    desc = "Enable EnhaPrio for elemental spec",
					    get = "GetProperty",
					    set = "SetProperty",
					    order = 2
					}
				}
			
			},
			visibilityGroup = {
			    type = "group",
			    name = "Visibility",
			    order = 2,
			    inline = true,
			    args = {
			        alwaysShow = {
				        type = "select",
				        name = "Always show",
				        desc = "Do you want the icons to show always or just when engaging an enemy.",
				        get = "GetProperty",
						set = "SetProperty",
						style = "dropdown",
						width = "double",
						values = {
							NO = "No, only when engaging an enemy",
							YES = "Yes, show everything always",
							BUFFS = "Yes, but buffs only",
						},
				    },
			    }
			},
			iconGroup = {
			    type = "group",
			    name = "Icon Settings",
			    order = 3,
			    inline = true,
			    args = {
			        size = {
						type = "range",
						name = "Main icon Size",
						desc = "Size of the icons",
						min = 1,
						max = 200,
						step = 1,
						get = "GetProperty",
						set = "SetProperty",
					},
					spacing = {
						type = 'range',
						name = "Icon Spacing",
						desc = "The distance between the displayed icons.",
						min = 0,
						max = 100,
						step = 1,
						get = "GetProperty",
						set = "SetProperty",
					},
					sizeFactor = {
						type = 'range',
						name = "Queued icon size",
						desc = "The size of the queued icons as a factor of the main icon.",
						min = 0.5,
						max = 1,
						step = 0.1,
						get = "GetProperty",
						set = "SetProperty",
					},
			    }
			},
			queueGroup = {
			    type = "group",
			    name = "Queue Settings",
			    order = 4,
			    inline = true,
			    args = {
			        queueDirection = {
				        type = "select",
				        name = "Queue Direction",
				        desc = "Which direction does the queue 'tail' point to.",
				        get = "GetProperty",
						set = "SetProperty",
						style = "dropdown",
						values = {
							RIGHT = "Right",
							LEFT = "Left",
							UP = "Up",
							DOWN = "Down"
						},
				    },
					maxQueue = {
					   	type = 'range',
						name = "Queue Size",
						desc = "Maximum number of spells shown in the queue.",
						min = 1,
						max = 8,
						step = 1,
						get = "GetProperty",
						set = "SetProperty",
					}
			    }
			},
			otherGroup = {
			    type = "group",
			    name = "Other Settings",
			    order = 5,
			    inline = true,
			    args = {
				updateFrequency = {
				    type = 'range',
				    name = "Update Frequency",
				    desc = "The delay in seconds to wait before rechecking the next spell.",
				    min = .05,
				    max = .5,
				    step = .05,
				    get = "GetProperty",
				    set = "SetProperty",
				    order = 2
				},
				displayMW = {
				    type = 'toggle',
				    name = "Show MW Tracker",
				    desc = "Show or hide the Maelstrom Weapon buff tracker.",
				    get = "GetProperty",
				    set = "SetProperty",
				    order = 3
				},
				displayGCD = {
				    type = 'toggle',
				    name = "Show GCD",
				    desc = "Show or hide global cooldown on the first icon.",
				    get = "GetProperty",
				    set = "SetProperty",
				    order = 4
				},
				enableLongCD = {
				    type = 'toggle',
				    name = "Track Long CD Spells",
				    desc = "Show or hide small icons for Feral Spirit and Fire Elemental Totem.",
				    get = "GetProperty",
				    set = "SetProperty",
				    order = 5
				},

				healthLevel = {
				    type = 'range',
				    name = "Critical Health Level",
				    desc = "When your health goes below this level, a heal will be suggested. Set to 0 to disable.",
				    min = 0,
				    max = 100,
				    step = 5,
				    get = "GetProperty",
				    set = "SetProperty",
				    order = 2
				},
			    }
			}
		}
	}
	return options
end
